#include "SerialCommand.h"
#include "EEPROMAnything.h"

#define SerialPort Serial1
#define Baudrate 9600

const int motorLeftForward = 3 ;
const int motorLeftBackward = 5 ;

const int motorRightForward = 9;
const int motorRightBackward = 6;

SerialCommand sCmd(SerialPort);

bool debug = false;
int run = 0 ;

unsigned long previous;
unsigned long calculationTime;

const int sensor[] = {A0, A1, A2, A3, A4, A5};

long normalised[6];
float debugPosition;

struct param_t
{
  unsigned long cycleTime;
  int black[6];
  int white[6];
  int power;
  float diff;
  float kp;
} params;

void setup()
{
  SerialPort.begin(Baudrate);
  sCmd.setDefaultHandler(onUnknownCommand);
  sCmd.addCommand("set", onSet);
  sCmd.addCommand("calibrate", onCalibrate);
  sCmd.addCommand("debug", onDebug);

  EEPROM_readAnything(0, params);
 
  SerialPort.println("ready");

  analogWrite(motorLeftForward, 0);
  analogWrite(motorLeftBackward, 00);
  analogWrite(motorRightBackward, 0);
  analogWrite(motorRightForward, 0);

}

void loop()
{
  sCmd.readSerial();
  unsigned long current = micros();
  if (current - previous > params.cycleTime){
    previous = current ;

    //normalisatie
    for (int i = 0; i < 6; i++)
    {
      long value = analogRead(sensor[i]);
      normalised[i] = map(value, params.black[i], params.white[i], 0, 1000);
    }

    //zwartste sensor zoeken
    int index = 0 ;
    for (int i = 0 ; i < 6 ; i++){
      if (normalised[i] < normalised[index]){
        index = i;
      }
    }

    //positie bepalen
    float position;
    if (index == 0){
      position = -30;
      debugPosition = position ;
    }
    else if (index == 0){
      position = 30;
      debugPosition = position ;
    }
    else {
      int sNul = normalised[index];
      int sMinEen = normalised[index - 1];
      int sPlusEen = normalised[index + 1];

      float b = (sPlusEen - sMinEen);
      b = b / 2;

      float a = sPlusEen - b - sNul ;

      position = -b / (2 * a) ; 
      position += index ;
      position -= 2.5;

      position *= 15;
      debugPosition = position ;
    }

    float error = -position;
    float output = error * params.kp ;

    output = constrain(output, -510, 510);

    int powerLeft = 0 ;
    int powerRight = 0 ;

    if (run == 1)
    {
      if (output >= 0)
      {
        powerLeft = constrain(params.power + params.diff * output, -255, 255);
        powerRight = constrain(powerLeft - output, -255, 255);
        powerLeft = powerRight + output;
      }
      else
      {
        powerRight = constrain(params.power - params.diff * output, -255, 255);
        powerLeft = constrain(powerRight + output, -255, 255);
        powerRight = powerLeft - output;
      }
    }
    analogWrite(motorLeftForward, powerLeft > 0 ? powerLeft : 0);
    analogWrite(motorLeftBackward, powerLeft < 0 ? powerLeft : 0);
    analogWrite(motorRightForward, powerRight > 0 ? powerRight : 0);
    analogWrite(motorRightBackward, powerRight < 0 ? powerRight : 0);
  }
  unsigned long difference = micros() - current;
  if (difference > calculationTime) calculationTime = difference;
}

void onSet(){
  char* param = sCmd.next();
  char* value = sCmd.next();
  
  if (strcmp(param, "cycle") == 0) {
    params.cycleTime = atol(value);
    SerialPort.print("CycleTime set to: ");
    SerialPort.print(params.cycleTime);
    SerialPort.println();
  }
  else if (strcmp(param, "power") == 0){
    params.power = atol(value);
    SerialPort.print("Power set to: ");
    SerialPort.println(params.power);
  }
  else if (strcmp(param, "diff") == 0){
    params.diff = atof(value);
    SerialPort.print("Diff set to: ");
    SerialPort.println(params.diff);
  }
  else if (strcmp(param, "kp") == 0){
    params.kp = atof(value);
    SerialPort.print("Kp set to: ");
    SerialPort.println(params.kp);
  }
  else if (strcmp(param, "run") == 0){
    run = atol(value);
    SerialPort.print("Run set to: ");
    SerialPort.println(run);
  }
  EEPROM_writeAnything(0, params);
}

void onDebug(){
  SerialPort.print("CycleTime set to: ");
  SerialPort.println(params.cycleTime);

  SerialPort.print("Run set to: ");
  SerialPort.println(run);

  SerialPort.print("Black: ");
  for (int i = 0 ; i < 6 ; i++) {
        SerialPort.print(params.black[i]);
        SerialPort.print("    ");
      }
  SerialPort.println();
      
  SerialPort.print("White: ");
  for (int i = 0 ; i < 6 ; i++) {
      SerialPort.print(params.white[i]);
      SerialPort.print("    ");
    }
  SerialPort.println();

  SerialPort.print("Values: ");
  for (int i = 0 ; i < 6 ; i++)
  {
    SerialPort.print(normalised[i]);
    SerialPort.print("    ");
  }
  SerialPort.println();

  SerialPort.print("Position: ");
  SerialPort.println(debugPosition);

  SerialPort.print("Power: ");
  SerialPort.println(params.power);

  SerialPort.print("Diff: ");
  SerialPort.println(params.diff);

  SerialPort.print("Kp: ");
  SerialPort.println(params.kp);

  SerialPort.print("Calculation Time: ");
  SerialPort.println(calculationTime);
}

void onCalibrate()
{
  char* param = sCmd.next();

  if (strcmp(param, "black") == 0){
    SerialPort.println("start calibrating black... ");
    for (int i = 0; i < 6; i++){
      int average = 0;
      average = gemiddeldeBepalen(i);
      params.black[i]= average ;
      SerialPort.print(average);
      SerialPort.print("    ");
    }
    SerialPort.println("done");
  }
  else if (strcmp(param, "white") == 0)
  {
    SerialPort.println("start calibrating white... ");    
    for (int i = 0; i < 6; i++){
      int average = 0;
      average = gemiddeldeBepalen(i);
      params.white[i]= average ;
      SerialPort.print(average);
      SerialPort.print("    ");
    }
    SerialPort.println("done");      
  }

  EEPROM_writeAnything(0, params);
}

int gemiddeldeBepalen(int sensorNummer){
  int totaal = 0 ;
  int gemiddelde = 0 ;
  for (int j = 0 ; j < 10 ; j++){
    totaal = totaal + analogRead(sensor[sensorNummer]);
  }
  gemiddelde = totaal / 10 ;
  return gemiddelde ;
}

void onUnknownCommand(char* command){
  SerialPort.print("Unknown Command: \"");
  SerialPort.print(command);
  SerialPort.println("\""); 
}
